/* dumpmem - a small utility to execute a binary and dump its memory */ 
/* contents before it begins executing.                              */ 
/*                                                                   */ 
/* Muli Ben-Yehuda <mulix@mulix.org>                                 */ 
/* Licensed under the GPL V2                                         */ 
/* $Id$                                                              */ 

#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h> 
#include <unistd.h> 
#include <sys/types.h> 
#include <errno.h> 
#include <sys/ptrace.h> 
#include <libiberty.h> 
#include <stdio.h> 
#include <string.h> 
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/param.h> 
#include <linux/user.h>     /* For constants ORIG_EAX etc */
#include <sys/syscall.h>    /* For SYS_write etc */

#include "util.h" 

const int word_size = sizeof(long); 

/* a linked list of process vmas */ 
struct map {
	struct map* next; 
	unsigned long begin; 
	unsigned long end; 
}; 

/* make a ptrace request and barf if it fails */ 
static unsigned long xptrace(enum __ptrace_request req, pid_t proc, void* addr)
{
	unsigned long res; 

	/* global errno */ 
	errno = 0; 
	res = ptrace(req, proc, addr, NULL); 
	if (res == (unsigned long)-1 && errno)
		fatal("ptrace(proc %d, req %d): %s", proc, req, strerror(errno)); 

	return res; 
}

/* get the data at 'addr */ 
static unsigned long getmem(pid_t proc, unsigned long addr)
{
	return xptrace(PTRACE_PEEKDATA, proc, (void*)addr); 
}

/* map constructor */ 
static struct map* make_map(unsigned long begin, unsigned long end)
{
	struct map* m = malloc(sizeof(*m)); 

	if (!m)
		fatal("malloc"); 

	m->begin = begin; m->end = end; 
	m->next = NULL; 
	return m; 
}

/* get the maps for the process */ 
static void get_maps(pid_t proc, struct map** m)
{
	/* open the maps file for this pid */ 
	char name[32]; 
	char buf[32000]; 
	int fd; 
	int ret; 
	int available = sizeof(buf); 
	struct map* head = NULL, *curr; 
	char* b, *e; 
	unsigned long begin, end; 

	(void)snprintf(name, sizeof(name), "/proc/%d/maps", (int)proc); 

	if (!(fd = open(name, O_RDONLY)))
		fatal("open %s: %s", name, strerror(errno)); 

	/* we make a single read and hope to read everything. Shaky at best */ 
	if ((ret = read(fd, buf, available - 1)) < 0)
		fatal("read %s: %s", name, strerror(errno)); 

	buf[ret] = '\0'; 

	close(fd); 
	/* we've read 'ret' bytes, parse it. file format: */ 
/* 
   08048000-08053000 r-xp 00000000 03:06 352128     /bin/ls
   08053000-08054000 rw-p 0000b000 03:06 352128     /bin/ls
   00000000-40013000 r-xp 00000000 03:06 512051     /lib/ld-2.2.5.so
   40013000-40014000 rw-p 00013000 03:06 512051     /lib/ld-2.2.5.so
   bffff000-c0000000 rwxp 00000000 00:00 0
*/ 

	/* this parser is ugly. rewrite */ 
	b = buf; 
	do { 
		/* find the first '-' */ 
		if (!(e = strchr(b, '-')))
			fatal("invalid maps file format!"); 
		
		*e = '\0'; 
		begin = strtoul(b, NULL, 16); 
		if (!*(b = e + 1))
			fatal("invalid maps file format!"); 
		
		if (!(e = strchr(b, ' ')))
			fatal("invalid maps file format!"); 
		
		*e = '\0'; 
		end = strtoul(b, NULL, 16); 

		/* make the map */ 
		if (!head)
			curr = head = make_map(begin, end); 
		else {
			curr->next = make_map(begin, end); 
			curr = curr->next; 
		}
		
		/* look for a \n and another line */ 
		if (!(b = strchr(e + 1, '\n')))
			break; 
		/* if we found a NULL */ 
		if (!*++b)
			break; 
	} while (1); 
	
	*m = head; 
}

/* dump the vma described by this map */ 
static void dump_vma(pid_t proc, struct map* m)
{
	unsigned long addr = m->begin; 
	int counter = 0; 
	printf("%08lx: ", addr); 
	do {
		unsigned long dat; 
		if (++counter == 7) {
			printf("\n%08lx: ", addr); 
			counter = 0; 
		}

		dat = getmem(proc, addr); 

		addr = addr + word_size; 

		printf("%08lx ", dat); 

	} while (addr < m->end); 

	printf("\n"); 
}

/* dump the entire memory for this process */ 
static void dump_proc_mem(pid_t proc)
{
	struct map *p, *del; 
	get_maps(proc, &p); 

	while (p) {
		printf("\n\nmap: 0x%08lx-0x%08lx\n\n", p->begin, p->end); 
		
		dump_vma(proc, p); 

		del = p; 
		p = p->next; 
		
		free(del); 
	}
}

/* copied from syscalltrack's sctrace/main.cpp. I can do that, I wrote it */ 
static char* find_in_path(const char* name)
{
	char* ret = NULL;
	char* env = getenv("PATH"); 
	char* beg, *end; 
	char buf[MAXPATHLEN + 256]; 

	if (!env)
		return NULL; 

	if (!(env = strdup(env)))
		return NULL; 

	beg = env; 

	/* find the first : character. */
	while (1) {
		end = strchr(beg, ':'); 
		if (end)
			*end = '\0'; 

		(void)snprintf(buf, sizeof(buf), "%s/%s", beg, name); 

		if (!access(buf, X_OK)) {
			ret = strdup(buf); 
			break; 
		}
		if (!end)
			break; 
		++end; 
		beg = end; 
	}
	free(env); 
	return ret; 
}

void exec_child(const char* exec, char** argv)
{
	const char* fullpath; 

	xptrace(PTRACE_TRACEME, 0, NULL); 

	if (*exec == '/')
		fullpath = exec; 
	else if ((*exec == '.') && 
		 ((strlen(exec) > 1) && (exec[1] == '.' || exec[1] == '/')))
		fullpath = exec; 
	else if (!(fullpath = find_in_path(exec)))
		fatal("could not find file to execute '%s'\n", exec); 

	printf("starting"); 
	execv(fullpath, argv); 
	fatal("execv: %s", strerror(errno)); 
}

static void run_tracer(pid_t child)
{
	int status; 
	unsigned long orig_eax; 
	int insyscall;
	int res; 

	res = waitpid(child, &status, 0);
	if (res != child)
		fatal("waitpid"); 
	
	if (WIFEXITED(status))
		return; 

	orig_eax = xptrace(PTRACE_PEEKUSER, child, (void*)(4 * ORIG_EAX)); 
	if (orig_eax == SYS_execve) {
		if(!insyscall) {    /* syscall entry */
			insyscall = 1;
			xptrace(PTRACE_SYSCALL, child, NULL); 
		} else { /* syscall exit, maps set, do our thing */ 
			dump_proc_mem(child); 
			xptrace(PTRACE_KILL, child, NULL); 
		}
	}
	
	do {
		res = waitpid(child, &status, 0); 
		if (res != child)
			fatal("waitpid: %s", strerror(errno)); 
	} while (!WIFEXITED(status) && !WIFSIGNALED(status)); 
}

/* main */ 
int main(int argc, char* argv[])
{
	pid_t child; 

	if (argc == 1)
		fatal("no program to run given"); 

	child = fork(); 
	if (child == -1)
		fatal("fork failed\n"); 

	if (child == 0) {
		/* child code */ 
		exec_child(argv[1], &argv[1]); 
	} else {
		run_tracer(child); 
	}

	return 0; 
}
